### Lab 3 - Live Forensics with Volatility part 2  


In this session  we'll be doing a bit more exploration with Volatility and see how we can use some features for further analysis. 

You will be useing the memory dump that you created when the Cryptominer was running for a few minutes. I'll also show how we can use the latest version of Volatility (that you cloned from Github) in this tutorial.  

Analysis begins with questions. They could be? 

	- What is this malware doing? 
	- What is it modifying? 
	- Are there any network activities? 
	- Did it create more process? 
	- Where is it hiding (if it has been deleted)

Let's explore the above	questions with our Memory dump from the server yesterday. 


Make sure that you're in the correct directory. 
- cd ~/memoryanalysis

1. What is the process id of krn and when was it launched? 

	- You can use Linux Plugin: linux_pslist , linux_pstree, linux_pidhashtable 
	- You may want to use "| less" or "| grep krn"
	- The Process ID is: <YOUR ANSWER HERE> 

```shell 
volatility  --plugins=. -f memUbuntu.raw --profile=LinuxUbuntu_4_4_0-21x64 \ 
linux_pidhashtable

```


2. Some plugins allows passing of paramaters (i.e. pid or file path). Let's find where _krn_ is executed from. 
	- We can use Linux Plugins: linux_proc_maps 
	- Change the PID to the process id of KRN in your memory dump 
	- krn execution path is: 

```
volatility --plugins=. -f memUbuntu.raw --profile=LinuxUbuntu_4_4_0-21x64 \
linux_proc_maps -p PID

```

3. Now we would like extract the file in memory, to do that we will need Inode information. 
	- We can use Linux Plugin: linux_find_file to find the Inode number & inode (in memory) for a file 


```
volatility  --plugins=. -f memUbuntu.raw --profile=LinuxUbuntu_4_4_0-21x64 \ 
linux_find_file -F "/home/analyst/workspace/malware/._lul/krn"

```

You should get something that looks like the following, take note of the Inode because we will need it. Your Inode number will be differen than the example below if you use your own memory capture. 


```

Volatility Foundation Volatility Framework 2.6.1
Inod

e Number                  Inode File Path
---------------- ------------------ ---------
            1333 0xffff88003d1004c8 /home/analyst/workspace/malware/._lul/krn

```


4. We can use Volatility to "dump" the file from memory using the information we have above. 
	- We will use the plugin linux_find_file and will pass the parameter -O $filelame
	- For example: 
	- Can you validate if the file is the same? 

```
volatility  --plugins=. -f memUbuntu.raw --profile=LinuxUbuntu_4_4_0-21x64 \ 
linux_find_file -i 0xffff88003d1004c8 -O dump.elf 
 
```


5. Another plugin we can use is linux_procdump 
	- We need to specify the folder for dumping the processes with -D 
	- If you don't pass a specific process id, if will try to dump everything. 
	- let's extract the binary from memory for our _krn_ process 

```
mkdir dump 
volatility  --plugins=. -f memUbuntu.raw --profile=LinuxUbuntu_4_4_0-21x64 \
linux_procdump -D dump -p PID 

```
	- can you check that the file is a binary? (hint: Use file)
	- the following is a sample output for the above command 
	- take note of your own process id 


```
xffff88003cf06e00 krn                  1351            0x0000000000400000 dump/krn.1351.0x400000

```

```
dump/krn.1351.0x400000:   ELF 64-bit LSB executable, x86-64, version 1 (GNU/Linux), \ 
dynamically linked, interpreter /lib64/ld-linux-x86-64.so.2, stripped

```



```

6. Can we see any network activities from _krn_?  
	- we can use the plugin linux_netstat 
	- What is the connection status? 
	- what other network information can we use? 



	```
volatility  --plugins=. -f memUbuntu.raw --profile=LinuxUbuntu_4_4_0-21x64 \
linux_netstat -p 1351

	```


7. Yara is an open source tool for malware researchers. It uses patter matching to identify and classify malware. Volatility has Yara integration and we can use it to scan the memory dump for signs of malware. 

```

volatility  --plugins=. -f memUbuntu.raw --profile=LinuxUbuntu_4_4_0-21x64 \ linux_yarascan  -y malware_rules.yar

```
- note the malware_rules.yar file contains signatures related to malware 
- Can you identify the malware using this technique? 
- this may take a while 


8. Discussion 
	- What do you like about Volatility? 
	- What are its limitation? 
	- How would you use this in a real scenario, what preparation would you have to make? 
	- What can be said about this cryptominer 
	- What threat information from our analysis can we use to mitigate further and how? 


9. Extra Credit 
	- Find out what linux_yarascan is for or what Yara is in general. 
	- How can we use Yara for this scenario? 


10. Resources
	- Check out the Volatility Foundation page 
	- Find CheatSheets 




eof. 
